/*

    	    	 Seek Transit
	
	         by John Walker
	    http://www.fourmilab.ch/
	    
       This program is in the public domain.

    Process the output of a DE118 solar system numerical
    integration, testing for transit events and solar system
    barycentre extrema and solar radius crossings.
    
    For details, please see: http://www.fourmilab.ch/documents/canon_transits/
 
*/

#include <stdio.h>
#include <math.h>

#include "readout.h"
#include "absbody.h"
#include "linalg.h"
#include "astro.h"
#include "julian.h"
#include "mathutil.h"

#define ELEMENTS(array) (sizeof(array)/sizeof((array)[0]))

static int seek_transit(abs_step *s, int outer, int inner, double *subtends, double *objang)
{
    Vector sun_outer, inner_outer, dir_cross,
    	   outer_sun, inner_sun;
    double sun_outer_dist, inner_angsin, sun_outer_hyp, sunrad_angsin;
    
    /*	Vectors from Sun to inner and outer objects.  */
    VecSub(s->obj[outer].pos, s->obj[Sun].pos, outer_sun);
    VecSub(s->obj[inner].pos, s->obj[Sun].pos, inner_sun);
    
    /*	Dot product of vectors from Sun to inner and outer
    	object is the projection of one of these vectors
	onto the other.  We care only about the sign of this
	projection, which indicates whether the angle between
	the objects is less than 180 degrees.  If so,
	we're looking at an inferior conjuction and hence
	possible transit.  Otherwise, this is a superior
	conjunction (objects on different sides of the Sun),
	and hence no transit is possible.  */
	
    if (VecDot(outer_sun, inner_sun) <= 0) {
    	*subtends = *objang = 0;    	    	/* Flag this as superior conjunction */
    	return 0;
    }
   
    /*	Vector from outer object to Sun.  */
    VecSub(s->obj[Sun].pos, s->obj[outer].pos, sun_outer);
    
    /*	Vector from outer object to inner object.  */
    VecSub(s->obj[inner].pos, s->obj[outer].pos, inner_outer);
    
    /*	Distance from outer object to Sun.  */
    sun_outer_dist = VecLen(sun_outer);
    
    /*	Now for a little trigonometry.  Consider the right triangle
    	with side a the vector from the outer object to the centre
	of the Sun and side b an orthogonal vector with magnitude
	the radius of the Sun.  The length of the hypotenuse is of course
	c = sqrt(a^2 + b ^2).  The sine of the radius vector observed
	from the outer object is then b / c.  */
    sun_outer_hyp = sqrt(sq(sun_outer_dist) + sq(SOLAR_RADIUS_AU));
    *subtends = sunrad_angsin = SOLAR_RADIUS_AU / sun_outer_hyp;
    
    /*	Normalise the direction vectors from the outer object
    	to the Sun and the inner object.  */
    VecNorm(sun_outer);
    VecNorm(inner_outer);
    
    /*	Now take the cross product of the two direction vectors.
    	Since we've just normalised them, the magnitude of this
	vector is the sine of the angle between them, which is
	the angle of displacement of the inner object from the
	centre of the Sun as seen from the outer object.  */
    VecCross(sun_outer, inner_outer, dir_cross);
    *objang = inner_angsin = VecLen(dir_cross);
    
    /*	If the object is in transit, then the absolute value of the
    	sine of its displacement is less than or equal to the
	absolute value of the sine of the angle subtended by the
	solar radius from the outer object.  Note that due to our
	performing the calculation with vectors, we needn't sweat
	the signs of the sines (always wanted to write that!).
	
	We do not worry about extreme grazes, the oblateness of the Sun,
	nor non object-centric viewpoints from the outer object.  These
	can be attended to, should the spirit so move us, with
	additional detail.  */
    
    return inner_angsin <= sunrad_angsin;
}

int main(int argc, char *argv[0])
{
    FILE *fi;
    abs_step st;
    static const int planidx[] = { Sun, Mercury, Venus, Earth, Moon,
    	    	    	    	   Mars, Jupiter, Saturn, Uranus, Neptune, Pluto };
    int foundlast = 0;
    double soff, barylast = 0, barylastjd = 0;
    int barylastsgn = 2, barylastcrosssgn = 2;

#define no_JTEST
#ifdef JTEST
    int y;    
    double tjd, ut, jdr;
    int yy, mon, dd, hh, mm, ss;
    
    for (y = -10000; y <= 10000; y += 1000) {
    	tjd = gregorian_to_jd(y, 5, 11);
	jd_to_gregorian(tjd, &yy, &mon, &dd);
    	printf("G %6d-05-11: %.2f   %d-%02d-%02d\n", y, tjd, yy, mon, dd);
	ut = jd_to_unix_time(tjd);
	jdr = unix_time_to_jd(ut);
	printf("U %6d-05-11: %.2f  =>  %.2f  =>  %.2f\n", y, tjd, ut, jdr);
    }
    
    for (y = -10000; y <= 10000; y += 1000) {
    	tjd = julian_to_jd(y, 5, 11);
	jd_to_julian(tjd, &yy, &mon, &dd);
    	printf("J %6d-05-11: %.2f   %d-%02d-%02d\n", y, tjd, yy, mon, dd);
	
    }

    for (tjd = 10.5; tjd >= -10.5; tjd--) {
	
	jd_to_gregorian(tjd, &yy, &mon, &dd);
	jd_to_civil_time(tjd, &hh, &mm, &ss);
	printf("JD %.4f  %d: %12s %d-%02d-%02d %02d:%02d\n",
	    tjd, jd_to_weekday(tjd), weekday_names[jd_to_weekday(tjd)], yy, mon, dd, hh, mm);
    }
    exit(0);
#endif

    if (argc != 2) {
    	printf("usage: transits output_file\n");
	return 2;
    }
    
    if ((fi = fopen(argv[1], "rb")) == NULL) {
    	printf("Cannot open input file %s\n", argv[1]);
	return 1;
    }
    
    while (readabs(fi, &st) == 0) {
    	double subt, obja;
    	int xin, xout;
	int foundthis = 0;

	for (xout = 2; xout < ELEMENTS(planidx); xout++) {
	    int plout = planidx[xout];
	    
	    for (xin = 1; xin < xout; xin++) {
	    	int plin = planidx[xin];
		
		/*  Ignore "transits" of Earth from the Moon.  Actually finding
		    lunar eclipses is a bit more complicated than what we're doing
		    here.  */
		    
		if (!((plout == Moon) && (plin == Earth))) {
		    if (seek_transit(&st, plout, plin, &subt, &obja)) {
			int yy, mon, dd, hh, mm, ss;

			jd_to_gregorian(st.jd, &yy, &mon, &dd);
			jd_to_civil_time(st.jd, &hh, &mm, &ss);
			printf("Transit of %-8s from %-8s %.4f  %12s %d-%02d-%02d %02d:%02d   %.4f  %.4f   %.4f\n",
		    	    abs_objname[plin], abs_objname[plout],
	    		    st.jd, weekday_names[jd_to_weekday(st.jd)], yy, mon, dd, hh, mm,
			    rtd(asin(subt)), rtd(asin(obja)), rtd(asin(subt)) - rtd(asin(obja)));
			foundthis++;
		    }
		}
	    }
	}
	
	/*  On the first step without a transit after having found one or more
	    transits on previous steps, flush the output buffer to the file.
	    This is purely for convenience, in that it permits monitoring the
	    run in real-time when the output is redirected or piped through
	    "tee".  */
	    
	if ((foundlast > 0) && (foundthis == 0)) {
	    fflush(stdout);
	}
	foundlast = foundthis;
	
	/*  Now we evaluate the position of the solar system barycentre
	    with respect to the Sun.  If it's at a maximum or minimum, or
	    passing through the notional surface of the Sun, log the event.  */
	
	soff = VecLen(st.obj[Sun].pos);
    	if ((barylastjd != 0) && (sgn(soff - barylast) != barylastsgn)) {
	    if (barylastsgn != 2) {
		int yy, mon, dd, hh, mm, ss;

		jd_to_gregorian(barylastjd, &yy, &mon, &dd);
		jd_to_civil_time(barylastjd, &hh, &mm, &ss);
		printf("Barycentre %s at %.4f %12s %d-%02d-%02d %02d:%02d (G = %.3f, %.2f%% of solar radius).\n",
	    		barylastsgn < 0 ? "minimum" : "maximum",
	    		barylastjd, weekday_names[jd_to_weekday(barylastjd)], yy, mon, dd, hh, mm,
			1 + ((barylast - SOLAR_RADIUS_AU) / SOLAR_RADIUS_AU),
			((barylast - SOLAR_RADIUS_AU) * 100) / SOLAR_RADIUS_AU);
	    }
	    barylastsgn = sgn(soff - barylast);
	}
	
    	if ((barylastjd != 0) && (sgn(soff - SOLAR_RADIUS_AU) != barylastcrosssgn)) {
	    if (barylastcrosssgn != 2) {
		int yy, mon, dd, hh, mm, ss;

		jd_to_gregorian(barylastjd, &yy, &mon, &dd);
		jd_to_civil_time(barylastjd, &hh, &mm, &ss);
		printf("Barycentre %s at %.4f %12s %d-%02d-%02d %02d:%02d (G = %.3f, %.2f%% of solar radius).\n",
	    		barylastcrosssgn < 0 ? "exiting" : "entering",
	    		barylastjd, weekday_names[jd_to_weekday(barylastjd)], yy, mon, dd, hh, mm,
			1 + ((barylast - SOLAR_RADIUS_AU) / SOLAR_RADIUS_AU),
			((barylast - SOLAR_RADIUS_AU) * 100) / SOLAR_RADIUS_AU);
	    }
	    barylastcrosssgn = sgn(soff - SOLAR_RADIUS_AU);
	}
	barylastjd = st.jd;
	barylast = soff;
    }

    fclose(fi);
    
    return 0;
}
